// SPDX-License-Identifier: MIT
// SPDX-FileCopyrightText: 2025 Loïc Le Page <llepage@igalia.com>

#include <iomanip>
#include <thread>

#include <libcamera/libcamera.h>

using namespace libcamera;
using namespace std::chrono_literals;

int main()
{
    logSetTarget(LoggingTargetNone);

    // Initialize the camera manager.
    auto camManager = std::make_unique<CameraManager>();
    camManager->start();

    // List cameras and select the first one.
    std::shared_ptr<Camera> selectedCamera;
    for (const auto& camera : camManager->cameras())
    {
        std::cout << "Camera found: " << camera->id() << std::endl;
        if (!selectedCamera)
        {
            selectedCamera = camera;
        }

        auto& propertiesList = camera->properties();
        auto idMap = propertiesList.idMap();

        std::cout << "# Properties:" << std::endl;
        for (const auto& [id, value] : propertiesList)
        {
            auto property = idMap->at(id);
            std::cout << "  " << property->name() << "(" << id << ") = " << value.toString() << std::endl;
        }

        std::cout << "# Controls:" << std::endl;
        for (const auto& [control, info] : camera->controls())
        {
            std::cout << "  " << control->name() << " = " << info.toString() << std::endl;
        }
    }

    if (!selectedCamera)
    {
        std::cerr << "No camera found" << std::endl;
        return -1;
    }

    // Lock the selected camera and choose a configuration for video display.
    selectedCamera->acquire();

    auto camConfig = selectedCamera->generateConfiguration({StreamRole::Viewfinder});
    if (camConfig->empty())
    {
        std::cerr << "No suitable configuration found for the selected camera" << std::endl;
        return -2;
    }

    auto& streamConfig = camConfig->at(0);
    std::cout << "Default camera configuration is: " << streamConfig.toString() << std::endl;

    streamConfig.size.width = 1920;
    streamConfig.size.height = 1080;
    streamConfig.pixelFormat = formats::RGB888;

    if (camConfig->validate() == CameraConfiguration::Invalid)
    {
        std::cerr << "Invalid camera configuration" << std::endl;
        return -3;
    }
    std::cout << "Targeted camera configuration is: " << streamConfig.toString() << std::endl;

    if (selectedCamera->configure(camConfig.get()) != 0)
    {
        std::cerr << "Failed to update the camera configuration" << std::endl;
        return -4;
    }
    std::cout << "Camera configured successfully" << std::endl;

    // Allocate the buffers pool used to fetch frames from the camera.
    Stream* stream = streamConfig.stream();
    auto frameAllocator = std::make_unique<FrameBufferAllocator>(selectedCamera);
    if (frameAllocator->allocate(stream) < 0)
    {
        std::cerr << "Failed to allocate buffers for the selected camera stream" << std::endl;
        return -5;
    }

    auto& buffersPool = frameAllocator->buffers(stream);
    std::cout << "Camera stream has a pool of " << buffersPool.size() << " buffers" << std::endl;

    // Create the requests used to fetch the actual camera frames.
    std::vector<std::unique_ptr<Request>> requests;
    for (auto& buffer : buffersPool)
    {
        auto request = selectedCamera->createRequest();
        if (!request)
        {
            std::cerr << "Failed to create a frame request for the selected camera" << std::endl;
            return -6;
        }

        if (request->addBuffer(stream, buffer.get()) != 0)
        {
            std::cerr << "Failed to add a buffer to the frame request" << std::endl;
            return -7;
        }

        requests.push_back(std::move(request));
    }

    // Connect the requests execution callback, it is called each time a frame
    // has been produced by the camera.
    selectedCamera->requestCompleted.connect(selectedCamera.get(), [&selectedCamera](Request* request) {
        if (request->status() == Request::RequestCancelled)
        {
            return;
        }

        // We can directly take the first request buffer as we are managing
        // only one stream. In case of multiple streams, we should iterate
        // over the BufferMap entries or access the buffer by stream pointer.
        auto buffer = request->buffers().begin()->second;
        auto& metadata = buffer->metadata();
        if (metadata.status == FrameMetadata::FrameSuccess)
        {
            // As we are using a RGB888 color format we have only one plane, but
            // in case of using a multiplanes color format (like YUV420) we
            // should iterate over all the planes.
            std::cout << "Frame #" << std::setw(2) << std::setfill('0') << metadata.sequence
                      << ": time=" << metadata.timestamp << "ns, size=" << metadata.planes().begin()->bytesused
                      << ", fd=" << buffer->planes().front().fd.get() << std::endl;
        }
        else
        {
            std::cerr << "Invalid frame received" << std::endl;
        }

        // Reuse the request buffer and re-queue the request.
        request->reuse(Request::ReuseBuffers);
        selectedCamera->queueRequest(request);
    });

    // Start the camera streaming loop and run it for a few seconds.
    selectedCamera->start();
    for (const auto& request : requests)
    {
        selectedCamera->queueRequest(request.get());
    }

    std::this_thread::sleep_for(1500ms);
    selectedCamera->stop();

    // Cleanup the resources. In fact those resources are automatically released
    // when the corresponding destructors are called. The only compulsory call
    // to make is selectedCamera->stop() as the camera streaming loop MUST be
    // stopped before releasing the associated buffers pool.
    frameAllocator.reset();
    selectedCamera->release();
    selectedCamera.reset();
    camManager->stop();

    return 0;
}
